const child_process = require("child_process")
  , EventEmitter = require("events")
  , path = require('path')
  , memHumanize = require('../../utilities/humanize-memory')
  , fse = require('fs-extra')
  , spawnProcessAsync = require('../../utilities/spawnProcessAsync')
  , BdpTaskAdapter = require("../BdpTaskAdapter");

class BdpDockerAdapter extends BdpTaskAdapter {
  constructor(opt) {
    opt.adapterName = "Docker";
    opt.adapterAuthor = "Chi Yang: chiyang1118@gmail.com";
    opt.dockerPath = opt.dockerPath || '/bin/docker';
    super(opt);
    this.dockerPath = this.options.dockerPath;
    this.options.stdoeMode = "pipe";
    this.options.tmpfs = this.options.tmpfs || undefined;
  }
  async beforeExit() {
    const taskIDs = Object.keys(this.runningTasks);
    while (taskIDs.length > 0) {
      const deletingTaskIDs = taskIDs.splice(0, 5);
      const deletingPromises = deletingTaskIDs.map((deletingTaskID) => {
        process.stderr.write("[" + new Date().toString() + "] Stop container: " + deletingTaskID + "\n");
        return spawnProcessAsync(this.dockerPath, ['stop', deletingTaskID], 'stop-container', {
          mode: 'pipe', verbose: false, shell: true
        }); 
      });
      await Promise.all(deletingPromises).catch(e => console.log(e));
    }
  }
  async processSpawn(taskID) {
    const taskEmitter = new EventEmitter();
    const taskObj = this.taskLogs[taskID];
    const RunningProcess = child_process.spawn(this.dockerPath, taskObj.args, {
      cwd: taskObj.option.cwd || this.options.cwd || process.cwd(),
      shell: true,
      env: process.env
    });
    RunningProcess.on("exit", (code, signal) => {
      return code == 0
        ? taskEmitter.emit("finish", 0, null)
        : taskEmitter.emit("error", code, signal);
    });
    taskObj.pid = RunningProcess.pid;
    return {
      jobID: RunningProcess.pid,
      stdoutStream: RunningProcess.stdout,
      stderrStream: RunningProcess.stderr,
      taskEmitter: taskEmitter
    };
  }
  async taskOverrides(taskObj) {
    taskObj.cpus = taskObj.option.cpus;
    const {value, unit} = memHumanize(taskObj.option.mem, 0);
    taskObj.mem = String(value) + unit.toLowerCase();
    if (!this.argTemplate) {
      this.argTemplate = await fse.readFile(path.join(__dirname, 'docker-arg-recipe.yaml'), "utf8");
    }
    taskObj = await super._parseRecipe(taskObj, this.argTemplate);
    taskObj.command = this.dockerPath + " " + taskObj.args.join(" ");
    return taskObj;
  }
}
module.exports = BdpDockerAdapter;
