const child_process = require('child_process')
    , os = require('os')
    , path = require('path')
    , fse = require('fs-extra')
    , escapeText = require('../_escape_text.js')
    , runProcess = require('../../utilities/runProcess')
    // , pathMapping = require('../../utilities/pathMapping')
    , EventEmitter = require('events')
    , BdpTaskAdapter = require("../BdpTaskAdapter");
class PbsEventEmitter extends EventEmitter {}
let canRun = true;
let isLeaving = false;
/**
 * TODO: 1. Allow setting downloading files or not
 * TODO: 2. When uploading files that are soft links that links to a path out of /share or /ProjectFolder, directly resolve the soft link and copy the files.
 * 
 */
const kubectlAsync = async function(options, kubectlArgs, useName, stdoutWriteStream, stderrWriteStream) {
    const args = [
        'run',
        '--rm'];
    if (useName) {
        args.push('--name');
        args.push(options.globalDockerName);
    }
    args.push.apply(args, [
        '--volumes-from', options.gcloudConfigContainer,
        '-v', options.kubeCredentialPath + '/config' + ':/root/.kube/config',
        'google/cloud-sdk', 'kubectl'
    ]);
    args.push.apply(args, kubectlArgs);
    return new Promise((resolve, reject) => {
        const returnObj = {
            stdout: '', stderr: '', error: null
        };
        // process.stderr.write('[command] ' + options.dockerPath + ' ' + args.join(' ') + '\n');
        const kubectlProcess = child_process.spawn(options.dockerPath, args, {shell: false});
        kubectlProcess.stdout.on('data', (data) => {
            if (data) {
                if (stdoutWriteStream && stdoutWriteStream.write) {
                    stdoutWriteStream.write(data);
                }else {
                    returnObj.stdout += data.toString();
                }
            }
        });
        kubectlProcess.stderr.on('data', (data) => {
            if (data) {
                if (stderrWriteStream && stderrWriteStream.write) {
                    stderrWriteStream.write(data);
                }else {
                    returnObj.stderr += data.toString();
                }
            }
        });
        kubectlProcess.on('error', (err) => {
            returnObj.error = err;
            reject(returnObj);
        });
        kubectlProcess.on('exit', (code) => {
            if (code === 0) {
                resolve(returnObj);
            }else {
                reject(returnObj);
            }
        });
    });
}

const createJobOverwrites = function(jobId, taskObj, pvcName) {
    
    const args = [taskObj.exec];
    args.push.apply(args, taskObj.args);
    const env = [{name: 'HOST_USER_ID', value: String(taskObj.option.uid)},
    {name: 'HOST_USER_GROUP', value: String(taskObj.option.gid)},
    {name: 'PACKAGE_FOLDER', value: String(taskObj.option.configFolder)},
    {name: 'PROJECT_FOLDER', value: String(taskObj.option.dataFolder)}];
    
    /*
    const envNames = Object.keys(process.env);
    for (const envName of envNames) {
        env.push({ name: envName, value: process.env[envName]});
    }
    */

    for (const envName of Object.keys(taskObj.option.env)) {
        env.push({ name: envName, value: taskObj.option.env[envName]});
    }
    const imagePullSecrets = [];
    if (taskObj.option['imagePullSecrets'] && Array.isArray(taskObj.option['imagePullSecrets'])) {
        imagePullSecrets.push.apply(imagePullSecrets,taskObj.option['imagePullSecrets'].map((secretName) => {
            return {name: secretName};
        }));
    }
    const projectPathObj = path.parse(taskObj.option.dataFolder);
    const job = {
        "apiVersion": "batch/v1",
        "kind": 'Job',
        "metadata": {
            name: jobId
        },
        "spec": {
            "backoffLimit": 1,
            "template": {
                "spec": {
                    "restartPolicy": 'Never',
                    "containers": [
                        {
                            "resources": {
                                requests: {cpu: taskObj.option.cpus, memory: taskObj.option.mem + 'i'}
                            }, 
                            "name": taskObj.name.toLowerCase(),
                            "image": taskObj.image,
                            "args": args,
                            "stdin": false,
                            "stdinOnce": false,
                            "tty": false,
                            "workingDir": taskObj.option.workdir || "/data",
                            "volumeMounts": [{
                                    "mountPath": "/data",
                                    "name": "job-store",
                                    "subPath": "data/" + projectPathObj.base
                                },{
                                    "mountPath": "/shared",
                                    "name": "job-store",
                                    "subPath": "data/shared"
                                }
                            ],
                            "env": env,
                            securityContext: taskObj.option.setUser? {
                                runAsUser: taskObj.option.uid,
                                fsGroup: taskObj.option.gid,
                                allowPrivilegeEscalation: false
                            } : {
                                allowPrivilegeEscalation: true
                            }
                        }
                    ],
                    imagePullSecrets: imagePullSecrets,
                    "volumes": [{
                        "name":"job-store",
                        "persistentVolumeClaim": {
                            claimName: pvcName
                        }
                    }]
                }
            }
        }
    }
    return JSON.stringify(job);
}

const rsyncData = async function(syncPath, options, direction) {
    // docker run --rm -v E:\bioDataProcessor\server\credentials\id_rsa:/config/id_rsa -v E:\bioDataProcessor\server\datasets:/data -e USER=bdpuser -e IP=35.230.45.107 -e REMOTE_MOUNT=~/bdp registry.gitlab.com/chiyang/bdp-rsync 5ac8208833bb47175829450e/
    direction = direction || 'upload'
    await runProcess(options.dockerPath, [
        'run', '--rm'
        , '--name', options.globalDockerName
        , '-v', options.remoteKeyFile + ':/config/id_rsa'
        , '-v', path.parse(options.dataFolder).dir + ':/data'
        , '-e', '\"USER=' + options.remoteUser + '\"'
        , '-e', '\"IP=' + options.remoteIP + '\"'
        , '-e', '\"REMOTE_MOUNT=' + options.remoteMountPoint + '\"',
        'registry.gitlab.com/chiyang/bdp-rsync'
        , direction, syncPath
    ], 'rsync-data');
}
/*
    const prepareTaskData = async function(taskObj, direction) {
    const taskOption = taskObj.nativeTask.option;
    const projectID = path.basename(taskOption.dataFolder);
    for (const eachArg of taskObj.nativeTask.args) {
        let relPath = path.relative(taskOption.dataFolder, String(eachArg));
        let syncPath = null;
        let dataFilePath = '';
        let syncSharedPath = null;
        if (relPath.indexOf('..') !== 0) {
            dataFilePath = eachArg;
        }else if (eachArg.indexOf('/data/') === 0) {
            relPath = path.relative('/data', eachArg);
            dataFilePath = path.join(taskOption.dataFolder, relPath);
        }
        const pathExists = await fse.pathExists(dataFilePath);
        if (!pathExists) {
            continue;
        }
        syncPath = projectID + '/' + relPath;
        const fileStats = await fse.lstat(dataFilePath);
        if (fileStats.isSymbolicLink()) {
            const realPath = await fse.realpath(dataFilePath);
            const relSharedPath = path.relative(path.resolve(taskOption.dataFolder, '../shared'), realPath);
            syncSharedPath = 'shared/' + relSharedPath;
        }
        if (syncPath) {
            await rsyncData(syncPath, taskOption, direction);
        }
        if (syncSharedPath) {
            await rsyncData(syncSharedPath, taskOption, direction);
        }
    }
}
*/

const createCluster = async function(options) {
    if (!canRun) { throw 'Process stopping';}
    await runProcess(options.dockerPath, [
        'run','--rm', '--name', options.globalDockerName, '--volumes-from', options.gcloudConfigContainer, 'google/cloud-sdk'
        , 'gcloud', 'config', 'set', 'project', options.project
    ], 'configure-project-name');
    if (!canRun) { throw 'Process stopping';}
    await runProcess(options.dockerPath, [
        'run','--rm', '--name', options.globalDockerName, '--volumes-from', options.gcloudConfigContainer, 'google/cloud-sdk'
        , 'gcloud', 'config', 'set', 'compute/zone', options.zone
    ], 'configure-compute-zone');
    if (!canRun) { throw 'Process stopping';}
    await runProcess(options.dockerPath, [
        'run', '--rm', '--name', options.globalDockerName, '--volumes-from', options.gcloudConfigContainer, 'google/cloud-sdk'
        , 'gcloud', 'container', 'clusters', 'create', options.clusterName, '--machine-type', options.machineType, '--num-nodes', options.clusterNodeNumber
        , '--scopes', options.scopes
    ], 'create-gke-cluster');
    if (!canRun) { throw 'Process stopping';}
}

class GKEAdapter extends BdpTaskAdapter {
    constructor (opt){
        /**
         * opt: {
         *  ...,
         *  project: skilful-bliss-193205,
         *  zone: 'us-west1-a',
         *  machineType: 'f1-micro',
         *  clusterNodeNumber: 2,
         *  scopes: 'compute-rw,storage-rw',
         *  gcloudConfigContainer: 'bdp-gcloud-config'
         *  clusterName: '',
         *  kubeCredentialPath: '',
         *  persistentVolumeClaim: '',
         *  resourceFolder: null, || configFolder + '/configs'
         *  resource: ['xxx.yaml', 'yyy.yaml'] // relative to config/configs path
         *  deleteClusterWhenNoJobs: true
         *  remoteMountPoint: '~/bdp'
         *  remoteUser: 'bdpuser'
         *  remoteIP: '35.230.45.107'
         *  remoteKeyFile: ''
         *  globalSync: true // if globalSync === true => exec bdp-rsync for the project folder before all tasks are processing
         *  closeAfterRun: false
         *  workdir: /data/{projectID}
         *  keepJobs: true
         * }
         */
        opt.adapterName = 'Google Kubernetes Engine';
        opt.adapterAuthor = 'Chi Yang: chiyang1118@gmail.com';
        super(opt);
        this.dockerPath = this.options.dockerPath || 'docker';
        this.options.clusterName = this.options.clusterName || 'bdp-gke-cluster';
        this.userInfo = os.userInfo();
        this.homedir = os.homedir();
        this.options.resourceFolder = this.options.resourceFolder || this.options.configFolder + '/configs';
        this.sparkJobID = null;
        this.sparkMaster = null;
        this.sparkStatusPath = null;
        const projectID = path.basename(this.options.dataFolder);
        this.options.globalDockerName = projectID + '-' + new Date().getTime() + '-' + Math.floor(1000000*Math.random());
        this.options.keepJobs = this.options.keepJobs === undefined ? true : (this.options.keepJobs ? true : false);
    }
    async beforeStart() {
        // Check if the cluster existed
        let isClusterExisted = false;
        canRun = true;
        isLeaving = false;
        console.log('Initiating the cluster.');
        try {
            const clusterListProc =  child_process.spawnSync(this.options.dockerPath, [
                'run', '--rm', '--name', this.options.globalDockerName, '--volumes-from', this.options.gcloudConfigContainer, 'google/cloud-sdk'
                , 'gcloud', 'container', 'clusters', 'list'
            ], {encoding: 'utf8'});
            console.log(clusterListProc.stdout);
            console.log(clusterListProc.stderr);
            clusterListProc.stdout.split('\n').forEach((line) => {
                const lineData = line.split(/\s+/);
                if (lineData[0] === this.options.clusterName) {
                    isClusterExisted = true;
                }
            });
        }catch(e) {
            console.log('Failed to create the cluster.');
            console.log(e);
            process.exit(1);
        }
        if (!canRun) { throw 'Process stopped.';}
        if (!isClusterExisted) {
            await createCluster(this.options);
        }
        if (!canRun) { throw 'Process stopped.';}
        await runProcess(this.options.dockerPath, [
            'run', '--rm', '--name', this.options.globalDockerName, '--volumes-from', this.options.gcloudConfigContainer
            , '-v', this.options.kubeCredentialPath + ':/root/.kube/', 'google/cloud-sdk', 'gcloud', 'container', 'clusters', 'get-credentials', this.options.clusterName, '--zone=' + this.options.zone
        ], 'get-gke-credentials');
        for (const resource of this.options.resources) {
            if (!canRun) { throw 'Process stopped.';}
            try {
                await runProcess(this.options.dockerPath, [
                    'run',
                    '--rm',
                    '--name', this.options.globalDockerName,
                    '--volumes-from', this.options.gcloudConfigContainer,
                    '-v', this.options.kubeCredentialPath + ':/root/.kube',
                    '-v', this.options.resourceFolder + ':/config',
                    'google/cloud-sdk', 'kubectl', 'create', '-f', '"' + escapeText('/config/' + resource) + '"'
                ], 'create-resources');
            }catch(e) {
                console.log(e);
            }
        }
        const projectID = path.basename(this.options.dataFolder);
        if (!canRun) { throw 'Process stopped.';}
        await rsyncData(projectID + '/', this.options, 'upload');
        if (!canRun) { throw 'Process stopped.';}
        await rsyncData('shared/', this.options, 'upload');
        if (!canRun) { throw 'Process stopped.';}
    }
    async logging(stdoe, processID) {
        const writeLogFS = fse.createWriteStream(stdoe);
        const kubectlResult = await kubectlAsync(this.options, ['logs', '-l=job-name=' + processID], false, writeLogFS);
        if (kubectlResult && kubectlResult.stderr) {
            process.stderr.write(kubectlResult.stderr);
        }
        return new Promise((resolve) => {
            writeLogFS.end(() => { resolve();});
        });
    };
    async beforeProcessExit (processID) {
        this.taskLogs[processID].end = new Date().valueOf();
        // console.log('Retrieving data for job ' + processID );
        // await prepareTaskData(this.taskLogs[processID], 'download');
        console.log('Logging job :' + processID );
        try {
            await this.logging(this.taskLogs[processID].stdout, processID);
        }catch(e) {
            console.log('Failed to log job');
            console.log(e);
        }
        if (this.taskLogs[processID].option && this.taskLogs[processID].option.keepJobs) {
            console.log('Removing job :' + processID);
            try {
                await kubectlAsync(this.options, ['delete', 'job', processID], false);
            }catch(e) {
                // console.log('Failed to remove job');
                // console.log(e);
            }
        }
        try {
            await this.updateStatus(true);
        }catch(e) {
            console.log('Failed to update job info');
            console.log(e);
        }
        delete this.runningTasks[processID];
        // const exitCode = await checkJobExitCode(jobID);
        // console.log('Synchronizing data for job ' + processID + ' ...');
        // await prepareTaskData(taskObj); 
        return;
    }
    async _runProcess(processID) {
        /**
         * 1. Check if the filePath (absolute) existed, ture => pathMapping and rsync
         * 2. Check if the filePath starts with /data/projectID, true => rsync
         */
        // if (!this.options.globalSync) {
        // console.log('Synchronizing data for job ' + processID + ' ...');
        if (!canRun) {
            console.log('Job ' + processID + ' will not be scheduled.');
            return Promise.reject();
        }
        const runningTasks = this.runningTasks;
        // await prepareTaskData(taskObj, 'upload');
        const taskLogs = this.taskLogs;
        const taskObj = taskLogs[processID];
        const option = taskObj.option;
        const cwd = option.cwd || process.cwd();
        const stdoutFile = path.resolve(option.taskLogFolder, processID + '-stdout.txt');
        // const stderrFile = path.resolve(this.options.taskLogFolder, processID + '-stderr.txt');
        const createJob = child_process.spawnSync(option.dockerPath, taskObj.args, {cwd: cwd, encoding: 'utf8'});
        runningTasks[processID] = new PbsEventEmitter();
        process.stderr.write(createJob.stderr);
        process.stdout.write(createJob.stdout);
        if (createJob.stderr) {
            throw 'Failed to create jobs\n';
        }
        const startTime = new Date().valueOf();
        let msg = '[' + new Date(startTime).toString() + '] ' + taskObj.taskName + ' Starting.';
        process.stderr.write(msg) + '\n';
        msg = '[command] ' + option.dockerPath + ' ' + taskObj.args.join(' ');
        process.stderr.write(msg + '\n');
        taskLogs[processID].pid = processID;
        taskLogs[processID].start = startTime;
        taskLogs[processID].command = option.dockerPath + ' ' + taskObj.args.join(' ');;
        taskLogs[processID].stdout = stdoutFile;
        taskLogs[processID].stderr = '';
        await this.updateStatus(true);
        return new Promise((resolve, reject) => {
            if (runningTasks[processID]) {
                runningTasks[processID].on('finish', () => {
                    this.beforeProcessExit(processID).then(() => {
                        taskLogs[processID].exitCode = 0;
                        resolve(0);
                    }).catch((err) => {
                        reject(err);
                    });
                });
                runningTasks[processID].on('error', () => {
                    this.beforeProcessExit(processID).then(() => {
                        taskLogs[processID].exitCode = 1;
                        reject(1);
                    }).catch((err) => {
                        reject(err);
                    });
                });
            }else {
                this.beforeProcessExit(processID).then(() => {
                    taskLogs[processID].exitCode = 1;
                    reject(1);
                }).catch((e) => {
                    reject(e);
                });
            }
        });
    }
    _addTasks(tasks) {
        for (const eachTask of tasks) {
            const taskOption = Object.assign({}, this.options, eachTask.option);
            eachTask.option = taskOption;
            let mem = taskOption.mem;
            mem = mem.toUpperCase().match(/^(\d+[EPTGMK])/);
            mem = mem ? mem[1]: '100M';
            eachTask.option.mem = mem;
            const processID = eachTask.name + '-' + new Date().getTime() + '-' + Math.floor(1000000*Math.random());
            let args = [
                'run', '--rm',
                '--volumes-from', taskOption.gcloudConfigContainer,
                '-v', taskOption.kubeCredentialPath + '/config' + ':/root/.kube/config',
                'google/cloud-sdk', 'kubectl', 'run', processID,
                '--image=' + eachTask.image, '--restart=OnFailure',
                '--overrides=' + createJobOverwrites(processID, eachTask, taskOption.persistentVolumeClaim)
            ];
            /*
            if (os.platform() !== 'win32') {
                args = args.map((arg) => {
                    return '"\'' + escapeText(arg) + '\'"';
                });
            }
            */
            this.taskLogs[processID] = {
                taskName: eachTask.name
                , pid: -1
                , image: eachTask.image
                , exec: eachTask.exec
                , args: args
                , command: null
                , start: null
                , end: null
                , stdout: ''
                , stderr: ''
                , nodes: taskOption.nodes
                , exitCode: null
                , taskArgs: eachTask.args
                , option: taskOption
            }
        }
    }
    async beforeExit() {
        console.log('Calling beforExit.');
        canRun = false;
        if (!isLeaving) {
            isLeaving = true;
            try {
                child_process.spawnSync(this.options.dockerPath, ['stop', this.options.globalDockerName], {shell: true, encoding: 'utf8'});
            }catch(e) {
                console.log(e);
            }
            const processIDs = Object.keys(this.runningTasks);
            if (processIDs.length > 0) {
                console.log('Stop running jobs ...');
            }
            for (const processID of processIDs) {
                if (this.runningTasks[processID] && this.runningTasks[processID].emit) {
                    console.log('Stopping job ' + processID);
                    try {
                        this.runningTasks[processID].emit('error');
                    }catch (e) {
                        try {
                            await this.beforeProcessExit(processID);
                            this.taskLogs[processID].exitCode = 1;
                        }catch(e) {console.log(e)};
                    }
                }
            }
        }
        try {
            await this.updateStatus(true);
        }catch(e) {
            console.log(e);
        }
        if (this.options.downloadAfterRun) {
            const projectID = path.basename(this.options.dataFolder);
            try {
                await rsyncData(projectID + '/', this.options, 'download');
                await rsyncData('shared/', this.options, 'download');
            }catch(e) {
                console.log(e);
            }
        }
        if (this.options.closeClusterAfterRun) {
            console.log('Checking if there are other jobs on the cluster ...');
            const checkJobs = await kubectlAsync(this.options, ['get', 'jobs', '-o', 'json'], false);
            let jobs = [];
            try {
                jobs = JSON.parse(checkJobs.stdout).items;
            }catch(e) {
                jobs = [];
            }
            if (jobs.length === 0) {
                console.log('Removing Cluster ...');
                // docker run -it --volumes-from bdp-gcloud-config google/cloud-sdk gcloudcontainer clusters delete bdp-gcloud-cluster --quiet
                child_process.spawnSync(this.options.dockerPath, [
                    'run', '--rm', '--volumes-from', this.options.gcloudConfigContainer, 'google/cloud-sdk'
                    , 'gcloud', 'container', 'clusters', 'delete', this.options.clusterName, '--quiet'
                ], {shell: false, encoding: 'utf8'});
                // process.stdout.write(removeCluster.stdout);
                // process.stderr.write(removeCluster.stderr);
            }
        }
    }
    async _checkStatus() {
        const runningTasks = this.runningTasks;
        const jobIDs = Object.keys(this.runningTasks);
        const taskLogs = this.taskLogs;
        const checkJobProcess = await kubectlAsync(this.options, ['get', 'jobs', '-o', 'json'], false);
        const jobs = {};
        const jobItems = JSON.parse(checkJobProcess.stdout).items;
        for (const jobItem of jobItems) {
            if (jobItem.metadata && jobItem.metadata.name) {
                jobs[jobItem.metadata.name] = jobItem;
            }
        }
        const currentStatus = {
            finished: 0,
            exit: 0,
            running: 0,
            queued: 0,
            others: []
        };
        for (const processID of Object.keys(runningTasks)) {
            if (jobs[processID]) {
                const status = jobs[processID].status;
                if (status.succeeded == 1) {
                    runningTasks[processID].emit('finish');
                    currentStatus.finished ++;
                }else if (status.failed >= 1) {
                    runningTasks[processID].emit('error');
                    currentStatus.exit ++;
                }else if (status.active === 1) {
                    currentStatus.running ++;
                }else {
                    console.log('Unexpected status encountered');
                    console.log(jobs[processID].status);
                }
            }else {
                try {
                    this.runningTasks[processID].emit('error');
                }catch(e) {
                    // console.log(e);
                }
            }
        }
        currentStatus.pending = Object.keys(taskLogs).filter((processID) => {
            return taskLogs[processID].exitCode === null && jobIDs.indexOf(taskLogs[processID].pid) < 0;
        }).length;
        currentStatus.done = Object.keys(taskLogs).filter((processID) => {
            return taskLogs[processID].exitCode !== null && jobIDs.indexOf(taskLogs[processID].pid) < 0;
        }).length;
        currentStatus.total = Object.keys(taskLogs).length;
        return currentStatus;
    }
}

module.exports = GKEAdapter;