const BdpDockerAdapter = require("../adapters/docker/docker_adapter.1"),
  BdpPbsAdapter = require("../adapters/pbs/pbs_adapter.1"),
  GKEAdapter = require("../adapters/gke/gke-adapter"),
  BdpLocalAdapter = require("../adapters/local/local-adapter"),
  fse = require("fs-extra"),
  path = require("path");
const _getAdapter = function(options) {
  let returnAdapter;
  switch (options.adapter) {
    case "pbs":
      returnAdapter = new BdpPbsAdapter(options).initialize();
      break;
    case "cwltool":
      returnAdapter = require("./lib/adapters/cwltool/cwltool_adapter")(
        options
      );
      break;
    case "gke":
      returnAdapter = new GKEAdapter(options).initialize();
      break;
    case "local":
      returnAdapter = new BdpLocalAdapter(options).initialize();
      break;
    default:
      returnAdapter = new BdpDockerAdapter(options).initialize();
      break;
  }
  return returnAdapter;
};
const getAdapter = function(options, configPath) {
  let configObj;
  options = options || {};
  const taskName = process.env.TASK_NAME || "default";
  const taskKey = process.env.TASK_KEY || "default";
  if (!options.adapterName) {
    options.adapterName = "Docker";
  }
  if (!options.adapterAuthor) {
    options.adapterAuthor = "Chi Yang";
  }
  if (!configPath) {
    if (process.env.PACKAGE_FOLDER) {
      configPath = path.join(
        process.env.PACKAGE_FOLDER,
        "./configs/package-config.json"
      );
    }
  }
  try {
    if (configPath) {
      configObj = fse.readJsonSync(
        path.resolve(process.env.PACKAGE_FOLDER, configPath)
      );
    }else {
      configObj = false;
    }
    if (configObj[taskKey]) {
      configObj = configObj[taskKey];
    } else if (configObj[taskName]) {
      configObj = configObj[taskName];
    } else {
      configObj = configObj["default"];
    }
  } catch (e) {
    configObj = false;
  }
  /** Will read the configObj to set options */
  if (!configObj) {
    configObj = {};
  }
  // hard-coded options should overwrite config object which can be altered from web pages.
  const usedOptions = Object.assign({}, configObj, options);
  if (!usedOptions.dockerPath) {
    usedOptions.dockerPath = process.env.DOCKER_PATH || "docker";
  }
  if (!usedOptions.cpus) {
    usedOptions.cpus = configObj.cpus || 1;
  }
  if (!usedOptions.mem) {
    usedOptions.mem = configObj.memory || "4g";
  }
  if (!usedOptions.concurrency) {
    usedOptions.concurrency = configObj.concurrency || 1;
  }
  if (options.debug) {
    console.log("The used runtime configs: ");
    console.log(usedOptions);
  }
  return _getAdapter(usedOptions);
};
module.exports = getAdapter;
