const child_process = require('child_process');
const fse = require('fs-extra');
const defaultOption = {
  stopSIGs: ['SIGTERM', 'SIGINIT'],
  verbose: false,
  mode: 'pipe', // pipe | memory | file
  stdoutPath: null,
  stderrPath: null,
  cwd: process.cwd(),
  shell: true,
  evn: process.env,
  encoding: 'utf8'
}
const endingWritableStreamAsync = (writableStream, lastData) => {
  return new Promise(resolve => {
    if (!writableStream || !writableStream.end) {
      resolve();
    }
    writableStream.on("finish", () => {
      resolve();
    });
    writableStream.end(lastData);
  });
};
module.exports = function(exec, args, processName, option){
  option = Object.assign({}, defaultOption, option);
  return new Promise((resolve, reject) => {
    if (option.verbose) {
      process.stderr.write('[' + new Date().toString() + '] ' + processName + ' Starting.\n');
      process.stderr.write('[command] ' + exec + ' ' + args.join(' ') + '\n');
    }
    try {
      const spawnedProcess = child_process.spawn(exec, args, {cwd: option.cwd, shell: option.shell});
      const stdoeObj = {stdout: '', stderr: ''}; // Used as string for mode === 'memory'; as writablestream for mode === 'file';
      if (option.mode === 'file') {
        if (option.stdoutPath) {
          stdoeObj.stdout = fse.createWriteStream(option.stdoutPath);
        }
        if (option.stderrPath) {
          stdoeObj.stderr = fse.createWriteStream(option.stderrPath);
        }
      }
      if (option.mode && option.mode !== 'ignore') {
        ['stdout', 'stderr'].forEach((stdoe) => {
          spawnedProcess[stdoe].on('data', data => {
            if (option.mode === 'pipe') {
              process[stdoe].write(data);
            }else if (option.mode === 'memory') {
              stdoeObj[stdoe] += data.toString();
            }else if (option.mode === 'file') {
              if (stdoeObj[stdoe] && stdoeObj[stdoe].write) {stdoeObj[stdoe].write(data);}
            }
          });
        });
      }
      spawnedProcess.on('error', (err) => {
        if (option.verbose) {
          process.stderr.write('[' + new Date().toString() + '] ' + processName + ' has errors.\n');
          process.stderr.write(JSON.stringify(err));
        }
        reject(err);
      });
      spawnedProcess.on('exit', (code, signal) => {
        if (option.verbose) {
          process.stderr.write('[' + new Date().toString() + '] ' + processName + ' Finished.\n');
        }
        endingWritableStreamAsync(stdoeObj.stdout)
        .then(() => {
          return endingWritableStreamAsync(stdoeObj.stderr);
        }).then(() => {
          resolve({
            exitCode: code
            , signal: signal
            , stdout: option.mode === 'memory' ? stdoeObj.stdout : null
            , stderr: option.mode === 'memory' ? stdoeObj.stderr : null
          });
        })
      });
      option.stopSIGs.forEach((stopSignal) => {
        spawnedProcess.on(stopSignal, () => {
          spawnedProcess.kill();
        });
      });
    }catch(e) {
      reject(e);
    }
  });
}
