const fse = require("fs-extra"),
  path = require("path"),
  adapters = {
    "pbs": "../adapters/pbs/pbs_adapter.1.js",
    "docker": "../adapters/docker/docker_adapter.1.js",
    "gke": "../adapters/gke/gke-adapter.js",
    "local": "../adapters/local/local-adapter.js",
    "remote-pbs": "../adapters/remote-pbs/remote-pbs-adapter.js"
  };
const _getAdapter = function(options) {
  if (!options.adapter) {
    throw "No adapter specified.";
  }
  if (!adapters[options.adapter]) {
    throw "The adapter " +
      options.adapter +
      " hasn't been registered. Didn't know how to initiate it.";
  }
  let TaskAdapter;
  try {
    /* TODO: [20181206] using remote JavaScript Module may be insecure! Try to use npm install some-other-adapter module to use!
     * I am considering to separte all the adapters: local, docker, (singularity), pbs, remote-pbs, etc, into different modules!
     * Some of them will be default adapters.
     */
    TaskAdapter = require(adapters[options.adapter]);
  } catch (e) {
    process.stdout.write(
      "Error loading the adapter, " +
        options.adapter +
        ", from the path: " +
        adapters[options.adapter] + '\n'
    );
    throw e;
  }
  const taskAdapter = new TaskAdapter(options);
  if (!taskAdapter.initialize) {
    throw "The adapter " +
      options.adapter +
      " does not have the initialize function.";
  }
  return taskAdapter.initialize();
};
const registerAdapter = function(name, path) {
  adapters[name] = path.resolve(process.env.PACKAGE_FOLDER, path);
};
const getAdapter = function(options, configPath) {
  let configObj;
  options = options || {};
  const taskName = process.env.TASK_NAME || "default";
  const taskKey = process.env.TASK_KEY || "default";
  if (!options.adapterName) {
    options.adapterName = "Docker";
  }
  if (!options.adapterAuthor) {
    options.adapterAuthor = "Chi Yang";
  }
  if (!configPath) {
    if (process.env.PACKAGE_FOLDER) {
      configPath = path.join(process.env.PACKAGE_FOLDER, "./configs/package-config.json");
    }
  }
  try {
    if (configPath) {
      configObj = fse.readJsonSync(path.resolve(process.env.PACKAGE_FOLDER, configPath));
    } else {
      configObj = false;
    }
    if (configObj[taskKey]) {
      configObj = configObj[taskKey];
    } else if (configObj[taskName]) {
      configObj = configObj[taskName];
    } else {
      configObj = configObj["default"];
    }
  } catch (e) {
    configObj = false;
  }
  /** Will read the configObj to set options */
  if (!configObj) {
    configObj = {};
  }
  // hard-coded options should overwrite config object which can be altered from web pages.
  const usedOptions = Object.assign({}, configObj, options);
  if (!usedOptions.dockerPath) {
    usedOptions.dockerPath = process.env.DOCKER_PATH || "docker";
  }
  if (!usedOptions.cpus) {
    usedOptions.cpus = configObj.cpus || 1;
  }
  if (!usedOptions.mem) {
    usedOptions.mem = configObj.memory || "4g";
  }
  if (!usedOptions.concurrency) {
    usedOptions.concurrency = configObj.concurrency || 1;
  }
  if (options.debug) {
    console.log("The used runtime configs: ");
    console.log(usedOptions);
  }
  return _getAdapter(usedOptions);
};
module.exports = {
  get: getAdapter,
  register: registerAdapter,
  getRegisteredAdapters: function() {
    return adapters;
  }
};
