const nunjucks = require("nunjucks"),
  yaml = require("js-yaml"),
  fse = require("fs-extra"),
  path = require("path"),
  url = require("url"),
  os = require("os"),
  createDefaultFilters = require("./createDefaultFilters.js"),
  taskAdapterProvider = require("./task-adapter-provider"),
  parseJson = require('json-parse-better-errors'),
  runProcessAsync = require("./runProcess");

const nunjuckEnv = new nunjucks.Environment([], {
  autoescape: false,
  trimBlocks: true,
  lstripBlocks: true
});
createDefaultFilters(nunjuckEnv);

const renderStringAsync = (templateString, templateData) => {
  return new Promise((resolve, reject) => {
    nunjuckEnv.renderString(templateString, templateData, (err, res) => {
      if (err) {
        return reject(err);
      }
      res = res.trim();
      resolve(res);
    });
  });
};
const argv = [];
for (let i = 3; i < process.argv.length; i++) {
  argv.push(process.argv[i]);
}
const taskName = argv[2];
const env = process.env;

const renderTaskObj = async renderedTask => {
  // renderedTask = renderedTask.replace(/\\{1,1}/g, `\\\\`);
  const taskObj = parseJson(renderedTask);
  /**
   * Check taskObj.args is a string or array
   */
  if (!Array.isArray(taskObj.args) && typeof taskObj.args === "string") {
    let args = [];
    try {
      args = yaml.safeLoad(taskObj.args);
    } catch (e) {
      console.log(e);
      throw "Your task argument is a string, but it's not a valid yaml format.";
    }
    if (Array.isArray(args)) {
      taskObj.args = args;
    } else {
      throw "Your task arguments are successfully resolved but it seems they are not an array.";
    }
  }
  return taskObj;
};
let variables = {};
const bdpTaskReaderAsync = async (taskContextFile, theAdapterInstance, taskGroupIndex) => {
  try {
    const template = await fse.readFile(taskContextFile, "utf8");
    const allTaskGroups = yaml.safeLoad(template);
    const allTasks = [];
    const taskGlobalOption = theAdapterInstance && theAdapterInstance.getOptions ? theAdapterInstance.getOptions() : {};
    if (taskGroupIndex >= 0 && taskGroupIndex < allTaskGroups.length) {
      const taskGroups = Array.isArray(allTaskGroups[taskGroupIndex]) ? allTaskGroups[taskGroupIndex] : [allTaskGroups[taskGroupIndex]];
      let renderedTask = "";
      for(const taskGroup of taskGroups) {
        if (taskGroup['var']) {
          const varNames = Object.keys(taskGroup['var']);
          const newVars = {};
          for (const varName of varNames) {
            if ( typeof taskGroup['var'][varName] === 'string') {
              let varString = taskGroup['var'][varName];
              const ifUsingDump = varString.match(/\|\s*dump\s*}}\s*$/);
              if (!ifUsingDump) {
                varString = varString.replace(/(^.*\S*\s*)}}\s*$/,'$1 | dump }}');
              }
              const resolvedVarString = await renderStringAsync(varString, {
                name: taskName,
                argv: argv,
                env: env,
                path: path,
                url: url,
                os: os,
                var: variables,
                option: taskGlobalOption
              });
              if (taskGlobalOption.debug) {
                console.log(`Resolved Variable: [${varString}]=>>>${resolvedVarString}<<<`);
              }
              if (resolvedVarString === '') {
                newVars[varName] = '';
              } else {
                newVars[varName] = parseJson(resolvedVarString);
              }
            }
          }
          variables = Object.assign({}, variables, newVars);
          if (taskGlobalOption.debug) {
            console.log('Resolved variables:');
            console.log(variables);
          }
        }
        if (taskGroup["batch-items"]) {
          if (taskGlobalOption.debug) {
            console.log("Reading batch-items:");
            console.log(taskGroup["batch-items"]);
          }
          let batchItemTemplate = taskGroup["batch-items"];
          const ifUsingDump = batchItemTemplate.match(/\|\s*dump\s*}}\s*$/);
          if (!ifUsingDump) {
            batchItemTemplate = batchItemTemplate.replace(/(^.*\S*\s*)}}\s*$/,'$1 | dump }}');
          }
          const items_json = await renderStringAsync(batchItemTemplate, {
            name: taskName,
            argv: argv,
            env: env,
            path: path,
            url: url,
            os: os,
            var: variables,
            option: taskGlobalOption
          });
          const items = parseJson(items_json);
          if (taskGlobalOption.debug) {
            console.log("Got items to execute:");
            console.log(items);
          }
          for (const item of items) {
            if (taskGlobalOption.debug) {
              console.log("Reading task-template:");
              console.log(taskGroup["task-template"]);
            }
            renderedTask = await renderStringAsync(
              JSON.stringify(taskGroup["task-template"]),
              {
                item: item,
                argv: argv,
                env: env,
                os: os,
                var: variables,
                path: path,
                url: url,
                option: taskGlobalOption
              }
            );
            if (taskGlobalOption.debug) {
              console.log("The resolved task template:");
              console.log(renderedTask);
            }
            if (os.platform() !== "win32") {
              renderedTask = renderedTask.replace(
                /(\$\{[a-zA-Z_][a-zA-Z0-9_]*?\})/g,
                "\\\\$&"
              );
            }
            const taskObj = await renderTaskObj(renderedTask);
            if (taskObj.condition === undefined || (taskObj.condition.toLowerCase && taskObj.condition.toLowerCase() === 'true')) {
              allTasks.push(taskObj);
            }
          }
        } else {
          if (taskGlobalOption.debug) {
            console.log("Reading task-template:");
            console.log(taskGroup["task-template"]);
          }
          renderedTask = await renderStringAsync(
            JSON.stringify(taskGroup["task-template"]),
            {
              argv: argv,
              env: env,
              os: os,
              path: path,
              url: url,
              var: variables,
              option: taskGlobalOption
            }
          );
          if (os.platform() !== "win32") {
            renderedTask = renderedTask.replace(
              /(\$\{[a-zA-Z_][a-zA-Z0-9_]*?\})/g,
              "\\\\$&"
            );
          }
          if (taskGlobalOption.debug) {
            console.log("The resolved task template:");
            console.log(renderedTask);
          }
          const taskObj = await renderTaskObj(renderedTask);
          if (taskObj.condition === undefined || (taskObj.condition.toLowerCase && taskObj.condition.toLowerCase() === 'true')) {
            allTasks.push(taskObj);
          }
        }
      }
      if (taskGlobalOption.preview) {
        console.log(allTasks);
        return [];
      }else {
        return allTasks;
      }
    } else {
      return null; // End of task iterator
    }
  } catch (e) {
    throw e;
  }
};

module.exports = {
  addFilter: function(filterName, func, isAsync) {
    nunjuckEnv.addFilter(filterName, func, isAsync);
    return this;
  },
  addAdapter: function(adapterName, adapterPath) {
    taskAdapterProvider.register(adapterName, adapterPath);
    return this;
  },
  readAsync: bdpTaskReaderAsync,
  readAndExecute: async (taskIndexFile, configPath) => {
    let isUsingYamlConfig = false;
    taskIndexFile = path.normalize(taskIndexFile);
    const argv = process.argv;
    const theTaskIndexPathExists = await fse.pathExists(taskIndexFile);
    if (theTaskIndexPathExists) {
      process.stdout.write("Reading task-index.yaml ...\n");
      const taskIndexYaml = await fse.readFile(taskIndexFile, "utf8");
      const taskIndexData = yaml.safeLoad(taskIndexYaml);
      const taskIndexOption = taskIndexData.option;
      const taskAdapters = taskIndexData.adapters;
      if (taskIndexData && taskIndexData.tasks && argv[2] && taskIndexData.tasks[argv[2]]) {
        isUsingYamlConfig = true;
        const theTaskYamlPath = path.resolve(
          path.parse(taskIndexFile).dir,
          taskIndexData.tasks[argv[2]].path || taskIndexData.tasks[argv[2]]
        );
        const taskYamlPathExist = await fse.pathExists(theTaskYamlPath);
        if (taskYamlPathExist) {
          if (taskAdapters) {
            const adapterNames = Object.keys(taskAdapters);
            for (const adapterName of adapterNames) {
              taskAdapterProvider.register(adapterName, taskAdapters[adapterName]);
            }
          }
          const usedAdapter = taskAdapterProvider.get(
            Object.assign({}, taskIndexOption, taskIndexData.tasks[argv[2]].option || {}),
            configPath
          );
          let taskIterator = 0;
          while (taskIterator >= 0) {
            const tasks = await bdpTaskReaderAsync(theTaskYamlPath, usedAdapter, taskIterator);
            if (tasks === null || !Array.isArray(tasks)) {
              taskIterator = -1;
              break;
            }
            if (tasks.length > 0) {
              await usedAdapter.execute(tasks);
            }
            taskIterator++;
          }
        } else {
          throw "The task definition file is not found: " + theTaskYamlPath;
        }
      } else {
        if (!taskIndexData.tasks) {
          process.stdout.write("Please check the task index format in the task-index.yaml.\n");
        } else if (argv[2] == null || argv[2] === "") {
          process.stdout.write("You have not define a task name.\n");
        } else if (!taskIndexData.tasks[argv[2]]) {
          process.stdout.write(`Cannot find the task name you specified: ${argv[2]}` + "\n");
        }
      }
    } else {
      process.stdout.write("No valid task-index.yaml file.");
    }
    /**
     *  if failed to read task-index.yaml or the tasks' yaml definition
     *  The following default will be used.
     */
    if (!isUsingYamlConfig) {
      if (process.env.DOCKER_IMG) {
        process.stdout.write(
          "Trying to execute this task inside the bdp-package: " +
            process.env.DOCKER_IMG +
            "\n"
        );
        const usedAdapter = taskAdapterProvider.get({}, configPath);
        const argv = process.argv;
        await usedAdapter.execute([
          {
            name: argv[2],
            image: process.env.DOCKER_IMG,
            exec: argv[2],
            args: argv.slice(3),
            option: {}
          }
        ]);
      } else {
        process.stdout.write(
          "No environment variable DOCKER_IMG (process.env.DOCKER_IMG).\nTrying to execute the task directly.\n"
        );
        await runProcessAsync(argv[2], argv.slice(3), argv[2]);
      }
    }
  }
};
