const walk = require('walk')
    , path = require('path');
module.exports = function(inputFolder, fileHookAsyncFunc) {
    const hasErrors = [];
    return new Promise((resolve, reject) => {
        const walker  = walk.walk(inputFolder, { followLinks: true });
        walker.on('errors', (root, errors , next) => {
            if (errors && errors.length > 0) {
                hasErrors.push.apply(hasErrors, errors);
            }
            next();
        });
        walker.on('file', (root, fileStats, next) => {
            const pathObj = path.parse(path.join(root, fileStats.name));
            fileHookAsyncFunc(pathObj, fileStats).then(() => {
                next();
            }).catch((err) => {
                hasErrors.push(err);
                next();
            });
        });
        walker.on('end', () => {
            if (hasErrors.length > 0) {
                reject(hasErrors);
            }else {
                resolve();
            }
        });
    });
}
